// Rtk09Labirint.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#pragma warning(disable: 4996)
#include <new>
#include <vector>
#include <cassert>
#include <algorithm>
#include "Rtk09Labirint.h"
#ifdef TEST_MY
#include "Rtk09LabirintMy.h"
#endif
using namespace std;

#ifdef _WIN32
#include <winsock.h>
#else
#include <sys/socket.h>
#endif
#include <process.h>

//-----------------------------------------------------------------------------
// TAngle
//-----------------------------------------------------------------------------

class TAngle
{
public:
	int x, y;
	TAngle(int x_, int y_) : x(x_), y(y_) { assert(x > 0); }
	int Cmp(const TAngle& a) const {
		int u = y * a.x - a.y * x;
		if (u < 0) return -1; else if (u > 0) return 1; else return 0; }
	bool operator < (const TAngle& a) const { return Cmp(a) < 0; }
	bool operator <= (const TAngle& a) const { return Cmp(a) <= 0; }
	bool operator > (const TAngle& a) const { return Cmp(a) > 0; }
	bool operator >= (const TAngle& a) const { return Cmp(a) >= 0; }
	bool operator == (const TAngle& a) const { return Cmp(a) == 0; }
	bool operator != (const TAngle& a) const { return Cmp(a) != 0; }
};

class TAngleRange
{
public:
	TAngle from, to;
	TAngleRange(TAngle from_, TAngle to_) : from(from_), to(to_) { }
};

//-----------------------------------------------------------------------------
// Non-public extensions of TVisibleTile
//-----------------------------------------------------------------------------

TAngle VisibleTileToAngle(const TVisibleTile& t) { return TAngle(t.rowsInFront, t.column); }
int VisibleTileDist2(const TVisibleTile& t) { return t.rowsInFront * t.rowsInFront + t.column * t.column; }

bool operator < (const TVisibleTile &s, const TVisibleTile &t)
{
	int c = VisibleTileToAngle(s).Cmp(VisibleTileToAngle(t));
	if (c == 0) c = VisibleTileDist2(s) - VisibleTileDist2(t);
	return c < 0;
}

//-----------------------------------------------------------------------------
// TBoard
//-----------------------------------------------------------------------------

typedef enum TTile_ { tileClear = 0, tileExit = 1, tileBlock = 2, tileBlockExit = 3, tileWall = 4 } TTile;

typedef enum TDir_ { dirRight = 0, dirUp = 1, dirLeft = 2, dirDown = 3 } TDir;
const int DX[4] = { 1, 0, -1, 0 };
const int DY[4] = { 0, -1, 0, 1 };

class TBoard
{
public:
	int w, h; // size of the playfield
	TTile *tiles;
	int xStart, yStart; TDir dirStart; // recommended starting position
	int vis; // recommended visibility

	TBoard() : w(0), h(0), tiles(0) { }
	TBoard(const char *fileName) : w(0), h(0), tiles(0) { Load(fileName); }
	TBoard(int w_, int h_, int vis_) : w(w_), h(h_), vis(vis_) {
		tiles = new TTile[w * h];
		assert(0 < w && w <= MaxW && 0 < h && h <= MaxH);
		for (int i = 0; i < w * h; i++) tiles[i] = tileClear; }
	TBoard(const TBoard& g) : w(g.w), h(g.h), xStart(g.xStart), yStart(g.yStart), dirStart(g.dirStart), vis(g.vis) {
		tiles = new TTile[w * h];
		for (int i = 0; i < w * h; i++) tiles[i] = g.tiles[i]; }
	~TBoard() { if (tiles) delete[] tiles; }
	TBoard& operator = (const TBoard &g) {
		if (&g != this) { this->~TBoard(); new (this) TBoard(g); }
		return *this; }
	void Load(const char *fileName);
	void Save(FILE *f) const;
	void Save(const char *fileName) const { FILE *f = fopen(fileName, "rt"); Save(f); fclose(f); }

	TTile GetTile(int x, int y) const {
		if (x < 0 || x >= w || y < 0 || y >= h) return tileClear;
		else return tiles[y * w + x]; }
	void SetTile(int x, int y, TTile t) {
		assert(0 <= x && x < w && 0 <= y && y < h);
		tiles[y * w + x] = t; }

	void BuildOuterWall() {
		for (int x = 0; x < w; x++) { SetTile(x, 0, tileWall); SetTile(x, h - 1, tileWall); }
		for (int y = 0; y < h; y++) { SetTile(0, y, tileWall); SetTile(w - 1, y, tileWall); } }

};

void TBoard::Save(FILE *f) const
{
	fprintf(f, "%d %d %d\n", w, h, vis);
	for (int y = 0; y < h; y++)
	{
		for (int x = 0; x < w; x++)
		{
			TTile tile = GetTile(x, y);
			char c = '.';
			if (tile == tileClear) c = '.';
			else if (tile == tileWall) c = (x == 0 || y == 0 || x == w - 1 || y == h - 1) ? 'Z' : '#';
			else if (tile == tileBlock) c = 'B';
			else if (tile == tileExit) c = 'x';
			else assert(false);
			if (x == xStart && y == yStart)
				c = (dirStart == dirLeft ? '<' : dirStart == dirRight ? '>' : dirStart == dirUp ? '^' : dirStart == dirDown ? 'v' : c);
			fputc(c, f);
		}
		fprintf(f, "\n");
	}
}

void TBoard::Load(const char *fileName)
{
	FILE *f = fopen(fileName, "rt"); assert(f);
	int ok = fscanf(f, "%d %d %d\n", &w, &h, &vis);
	assert(ok == 3);
	assert(0 < w && w <= MaxW && 0 < h && h <= MaxH);
	tiles = new TTile[w * h];
	xStart = -1; yStart = -1; dirStart = (TDir) -1;
	for (int y = 0; y < h; y++)
	{
		for (int x = 0; x < w; x++)
		{
			char c = fgetc(f);
			TTile tile = tileClear;
			if (c == 'Z' || c == 'z') { assert(x == 0 || x == w - 1 || y == 0 || y == h - 1); tile = tileWall; }
			else if (c == '#' || c == 'W' || c == 'w') tile = tileWall;
			else if (c == 'B' || c == 'b' || c == 'O') tile = tileBlock;
			else if (c == ':' || c == 'x' || c == 'X') tile = tileExit;
			else if (c == '.') tile = tileClear;
			else if (c == '<' || c == '^' || c == 'v' || c == '>') {
				assert(xStart < 0);
				xStart = x; yStart = y;
				dirStart = (c == '<' ? dirLeft : c == '^' ? dirUp : c == 'v' ? dirDown : dirRight);
				tile = tileClear; }
			else
				assert(false);
			SetTile(x, y, tile);
		}
		// Read the EOLN; allow whitespace before it.
		while (true) {
			int c = fgetc(f);
			if (c == EOF) { assert(y == h - 1); break; }
			if (c == '\n') break;
			assert(isspace((int) (unsigned char) c)); }
	}
	fclose(f);
}

//-----------------------------------------------------------------------------
// TPlayer
//-----------------------------------------------------------------------------
// Maintains the player's location, direction and line of sight.

typedef enum TSeen_ { seenNever = 0, seenBefore = 1, seenNow = 2 } TSeen;

class TMoveCounts
{
public:
	int nStepsWithBurden, nStepsWoBurden, nTurnsWithBurden, nTurnsWoBurden, nUTurnsWithBurden, nUTurnsWoBurden, nPickups, nPickupAttempts, nDrops;
	TMoveCounts() { Clr(); }
	void Clr() {
		nStepsWithBurden = 0; nStepsWoBurden = 0;
		nTurnsWithBurden = 0; nTurnsWoBurden = 0;
		nUTurnsWithBurden = 0; nUTurnsWoBurden = 0;
		nPickups = 0; nPickupAttempts = 0; nDrops = 0; }
	TMoveCounts& operator += (const TMoveCounts& mc) {
		nStepsWithBurden += mc.nStepsWithBurden;
		nStepsWoBurden += mc.nStepsWoBurden;
		nTurnsWithBurden += mc.nTurnsWithBurden;
		nTurnsWoBurden += mc.nTurnsWoBurden;
		nUTurnsWithBurden += mc.nUTurnsWithBurden;
		nUTurnsWoBurden += mc.nUTurnsWoBurden;
		nPickups += mc.nPickups;
		nPickupAttempts += mc.nPickupAttempts;
		nDrops += mc.nDrops;
		return *this; }
	int TotalCost() const { return
		nStepsWoBurden * MoveCost::WalkNB +
		nTurnsWoBurden * MoveCost::TurnNB +
		nUTurnsWoBurden * MoveCost::UTurnNB +
		nPickups * MoveCost::PickUp +
		nPickupAttempts * MoveCost::TryPickup +
		nDrops * MoveCost::Drop +
		nStepsWithBurden * MoveCost::WalkB +
		nTurnsWithBurden * MoveCost::TurnB +
		nUTurnsWithBurden * MoveCost::UTurnB; }
	int TotalMoves() const { return
		nStepsWoBurden +
		nTurnsWoBurden +
		nUTurnsWoBurden +
		nPickups +
		nPickupAttempts +
		nDrops +
		nStepsWithBurden +
		nTurnsWithBurden +
		nUTurnsWithBurden; }
};

class TPlayer
{
private:
	TPlayer(TPlayer& p) { assert(false); }
	TPlayer& operator = (const TPlayer& p) { assert(false); return *this; }

public:
	int x, y;
	TDir dir;
	bool carries;
	int vis; // visibility -- how many tiles ahead does the player see?
	int w, h, ws, hs;
	TMoveCounts mc;

	enum { verbose = true };

	TSeen *seen;
	TBoard *board;
	char *visArea;
	vector<TVisibleTile> visTiles;

	TPlayer(TBoard *board_, int vis_, int x_, int y_, TDir dir_)
		: board(board_), vis(vis_), x(x_), y(y_), dir(dir_)
	{
		assert(board); assert(vis > 0);
		w = board->w; h = board->h;
		ws = w + 2 * vis; hs = h + 2 * vis;
		seen = new TSeen[ws * hs];
		visArea = new char[(2 * vis + 1) * vis];
		Reset(x_, y_, dir_);
	}
	~TPlayer() { delete[] seen; delete[] visArea; }

	void Reset(int x_, int y_, TDir dir_)
	{
		x = x_; y = y_; dir = dir_; carries = false;
		assert(0 <= x && x < w && 0 <= y && y < h);
		TTile tile = board->GetTile(x, y); assert(tile == tileClear || tile == tileExit);
		for (int i = 0; i < ws * hs; i++) seen[i] = seenNever;
		for (int i = 0; i < (2 * vis + 1) * vis; i++) visArea[i] = visUnknown;
		// Make sure that the maze is walled all around, except for at least one exit.
		bool hasExits = false;
		for (int x = 0; x < w; x++) {
			tile = board->GetTile(x, 0); assert(tile == tileWall || tile == tileExit); if (tile == tileExit && x != 0 && x != w - 1) hasExits = true;
			tile = board->GetTile(x, h - 1); assert(tile == tileWall || tile == tileExit); if (tile == tileExit && x != 0 && x != w - 1) hasExits = true; }
		for (int y = 0; y < h; y++) {
			tile = board->GetTile(0, y); assert(tile == tileWall || tile == tileExit); if (tile == tileExit && y != 0 && y != h - 1) hasExits = true;
			tile = board->GetTile(w - 1, y); assert(tile == tileWall || tile == tileExit); if (tile == tileExit && y != 0 && y != h - 1) hasExits = true; }
		assert(hasExits);
		//
		mc.Clr();

		RefreshVisibleArea();
		ApplyVisAreaToSeen(true);
	}

	// Note: you should initialize the player with Reset (or with the constructor) before calling RunGame.
	void RunGame(IBehavior *engine, FILE *fLog)
	{
		if (fLog) { board->Save(fLog); fprintf(fLog, "PlayerName %s\n", engine->GetName().c_str()); fflush(fLog); }
		TTimer tm; tm.Start();
		engine->BeginGame(w, h, vis); tm.Stop();
		enum { MaxMoves = 100000 };
		bool won = false; string err; int nMoves = 0;
		for (nMoves = 0; nMoves < MaxMoves; nMoves++)
		{
			tm.Start();
			TMove move = engine->NextMove(carries, visTiles);
			tm.Stop();
			if (verbose) printf("Move %d: %s\n", nMoves, MoveStr(move));
			if (fLog) { fprintf(fLog, "Move %s\n", (move == moveForward ? "Forward" :
				move == moveTurnLeft ? "TurnLeft" :
				move == moveTurnRight ? "TurnRight" :
				move == moveUTurn ? "UTurn" :
				move == movePickUp ? "PickUp" :
				move == moveDrop ? "Drop" : "Invalid"));
				fflush(fLog); }
			bool ok = PerformMove(move, err);
			if (verbose) printf("New state: (%d, %d), dir %d; CPU time = %.2f sec\n", x, y, (int) dir, tm.Sec());
			if (! ok) break;
			if (board->GetTile(x, y) == tileExit) { won = true; nMoves++; break; }
		}
		if (won) err = "Cestitamo, prisel si do izhoda.";
		else if (nMoves > MaxMoves) err = "Naredil si prevec korakov.";
		tm.Start();
		engine->EndGame(won, err);
		tm.Stop();
		if (fLog)  { fprintf(fLog, "%s %s\n", (won ? "Won" : "Error"), err.c_str()); fflush(fLog); }
		FILE *f = stdout;
		fprintf(f, "RunGame report: board size %d * %d, visibility %d\n", w, h, vis);
		fprintf(f, "- %d steps w/ burden, %d steps w/o burden\n", mc.nStepsWithBurden, mc.nStepsWoBurden);
		fprintf(f, "- %d turns w/ burden, %d turns w/o burden\n", mc.nTurnsWithBurden, mc.nTurnsWoBurden);
		fprintf(f, "- %d u-turns w/ burden, %d u-turns w/o burden\n", mc.nUTurnsWithBurden, mc.nUTurnsWoBurden);
		fprintf(f, "- %d pickups, %d pickup attempts, %d drops\n", mc.nPickups, mc.nPickupAttempts, mc.nDrops);
		fprintf(f, "- Total number of moves: %d; totalCost = %d; CPU time = %.2f sec\n", nMoves, mc.TotalCost(), tm.Sec());
		fprintf(f, "- %s (%s)\n", won ? "SUCCESS!" : "FAILURE!", err.c_str());
	}

	TSeen GetSeen(int x, int y) const {
		if (x < -vis || x >= w + vis || y < -vis || y >= h + vis) return seenNever;
		else return seen[(y + vis) * ws + (x + vis)]; }
	void SetSeen(int x, int y, TSeen s) {
		assert(-vis <= x && x < w + vis && -vis <= y && y < h + vis);
		seen[(y + vis) * ws + (x + vis)] = s; }

	void PrintVisArea(FILE *f = 0);
	void PrintSeen(FILE *f = 0);

	bool PerformMove(TMove move, string &errorMsg);
protected:
	static void UpdateCoveredRanges(vector<TAngleRange> &cov, TAngle from, TAngle to, bool &wasFullyCovered);

	// Call with seenNow = false before a move and seenNow = true after a move (and a call to RefreshVisibleArea()).
	void ApplyVisAreaToSeen(bool seenNow);
	// 'visBuf' must point to a buffer of 'vis' rows and (2 * vis + 1) chars in each row.
	void RefreshVisibleArea();

	bool ApplyMove(TMove move, string &errorMsg);
};

void TPlayer::ApplyVisAreaToSeen(bool isSeenNow)
{
	int dxAway = DX[dir], dyAway = DY[dir];
	int dxRight = DX[(dir + 3) % 4], dyRight = DY[(dir + 3) % 4];
	for (int vy = 1; vy <= vis; vy++) for (int vx = -vis; vx <= vis; vx++)
	{
		char c = visArea[(vy - 1) * (2 * vis + 1) + (vx + vis)];
		if (c == visUnknown) continue;
		int xx = x + vy * dxAway + vx * dxRight;
		int yy = y + vy * dyAway + vx * dyRight;
		if (xx < -vis || xx >= w + vis || yy < -vis || yy >= h + vis)
			// The player is moving very far from the board -- we won't bother tracking
			// what he has or hasn't seen there.
			continue;
		// Test.
		TTile tile = board->GetTile(xx, yy);
		if (c == visOuterWall) assert(tile == tileWall && (xx == 0 || xx == w - 1 || yy == 0 || yy == h - 1));
		else if (c == visBlockOrWall) assert(tile == tileBlock || tile == tileWall || tile == tileBlockExit);
		else if (c == visExit) assert(tile == tileExit && (xx == 0 || xx == w - 1 || yy == 0 || yy == h - 1));
		else { assert(c == visClear); assert(tile == tileClear); }
		//
		if (isSeenNow) SetSeen(xx, yy, seenNow);
		else SetSeen(xx, yy, seenBefore);
	}
}

// On input, the ranges in 'cov' must be in strictly ascending order and must not touch.
void TPlayer::UpdateCoveredRanges(vector<TAngleRange> &cov, TAngle from, TAngle to, bool &wasFullyCovered)
{
	// Note that between any two adjacent ranges in 'cov' there is a strictly nonzero gap.
	// Thus, the only way for [from, to] to be fully covered is if it is fully covered
	// by an individual range in 'cov'.
	wasFullyCovered = false;
	vector<TAngleRange> cov2;
	size_t n = cov.size(), ri = 0;
	while (ri < n && cov[ri].from <= from) { cov2.push_back(cov[ri]); ri++; }
	if (cov2.size() == 0) cov2.push_back(TAngleRange(from, to)); // insert the new range at the beginning
	else {
		TAngleRange &last = cov2[cov2.size() - 1];
		assert(last.from <= from);
		if (to <= last.to) { wasFullyCovered = true; return; }
		if (from <= last.to) last.to = to; // extend 'last'
		else cov2.push_back(TAngleRange(from, to)); // insert the new range after 'last'
	}
	while (ri < n) {
		TAngleRange &last = cov2[cov2.size() - 1];
		if (last.to < cov[ri].from) cov2.push_back(cov[ri]); // append 'cov[ri]'
		else if (last.to < cov[ri].to) last.to = cov[ri].to; // extend 'last'
		ri++; }
	// Test.
	for (int i = 0; i < (int) cov2.size(); i++) {
		assert(cov2[i].from < cov2[i].to);
		if (i > 0) assert(cov2[i - 1].to < cov2[i].from); }
	cov = cov2;
}

void TPlayer::RefreshVisibleArea()
{
	int dxAway = DX[dir], dyAway = DY[dir];
	int dxRight = DX[(dir + 3) % 4], dyRight = DY[(dir + 3) % 4];
	vector<TAngleRange> cov; // covered ranges; angles are in strictly ascending order and the ranges do not touch
	visTiles.clear();
	// Cover everything outside a cone extending 45 degrees around the direction we're looking towards.
	cov.push_back(TAngleRange(TAngle(1, -(2 * h + 1)), TAngle(1, -1)));
	cov.push_back(TAngleRange(TAngle(1, 1), TAngle(1, 2 * h + 1)));

	for (int vy = 1; vy <= vis; vy++) for (int ivx = 0; ivx <= 2 * vis; ivx++)
	{
		int vx = ((ivx % 2) == 1 ? -1 : 1) * ((ivx + 1) / 2); // 0, 1, 2, 3, 4, ... -> 0, -1, 1, -2, 2, ...
		if (abs(vx) > vy) continue; // we see only 45 degrees left/right from our current direction
		TAngle from = vx <= 0 ? TAngle(2 * vy - 1, 2 * vx - 1) : TAngle(2 * vy + 1, 2 * vx - 1);
		TAngle to = vx < 0 ? TAngle(2 * vy + 1, 2 * vx + 1) : TAngle(2 * vy - 1, 2 * vx + 1);
		// Find out if this tile is fully covered by other blocks and walls.
		bool fullyCovered = false;
		for (int ri = 0; ri < (int) cov.size(); ri++)
			if (cov[ri].from <= from && to <= cov[ri].to) { fullyCovered = true; break; }
		//
		char c = visUnknown;
		if (! fullyCovered)
		{
			// We can report what's on this tile.
			int xx = x + vy * dxAway + vx * dxRight;
			int yy = y + vy * dyAway + vx * dyRight;
			TTile tile = board->GetTile(xx, yy);
			if (tile == tileClear) c = visClear;
			else if (tile == tileExit) c = visExit;
			else if (tile == tileWall && (xx == 0 || xx == w - 1 || yy == 0 || yy == h - 1)) c = visOuterWall;
			else if (tile == tileBlock || tile == tileBlockExit || tile == tileWall) c = visBlockOrWall;
			else assert(false); // unknown tile type?!
			if (c == visOuterWall || c == visBlockOrWall) { // This tile will cover some others.
				bool wasFullyCovered; UpdateCoveredRanges(cov, from, to, wasFullyCovered);
				assert(! wasFullyCovered); }
		}
		visArea[(vy - 1) * (2 * vis + 1) + (vx + vis)] = c;
		if (c != visUnknown)
			visTiles.push_back(TVisibleTile(vx, vy, c));
	}
	sort(visTiles.begin(), visTiles.end());
}

void TPlayer::PrintVisArea(FILE *f)
{
	if (! f) f = stdout;
	fprintf(f, "Visible area [x = %d, y = %d, dir = %s, visibility = %d]:\n", x, y, (dir == dirUp ? "N" : dir == dirLeft ? "W" : dir == dirDown ? "S" : dir == dirRight ? "E" : "?"), vis);
	for (int y = vis; y >= 0; y--) for (int x = -vis; x <= vis; x++)
	{
		char c;
		if (y == 0) c = (x == 0 ? 'P' : visUnknown);
		else c = visArea[(y - 1) * (2 * vis + 1) + (x + vis)];
		if (c == visUnknown) c = '+';
		if (x == -vis) fprintf(f, "    ");
		fputc(c, f);
		if (x == vis) fprintf(f, "\n");
	}
	fprintf(f, "%d visible tiles: ", (int) visTiles.size());
	for (int i = 0; i < (int) visTiles.size(); i++)
		fprintf(f, " %c(%d, %d)", visTiles[i].what, visTiles[i].column, visTiles[i].rowsInFront);
	fprintf(f, "\n");
}

void TPlayer::PrintSeen(FILE *f)
{
	if (! f) f = stdout;
	fprintf(f, "Board [w = %d, h = %d; visibility = %d]:\n", w, h, vis);
	for (int y = -vis; y < h + vis; y++)
	{
		fprintf(f, "    ");
		for (int x = -vis; x < w + vis; x++)
		{
			TTile tile = board->GetTile(x, y);
			TSeen seen = GetSeen(x, y);
			char c;
			if (seen == seenNow) c = (tile == tileWall ? '#' : tile == tileBlock ? 'O' : tile == tileBlockExit ? '0' : tile == tileExit ? 'x' : '_');
			else if (seen == seenBefore) c = (tile == tileWall ? 'W' : tile == tileBlock ? 'B' : tile == tileBlockExit ? 'Q' : tile == tileExit ? ';' : ',');
			else c = (tile == tileWall ? 'w' : tile == tileBlock ? 'b' : tile == tileBlockExit ? 'q' : tile == tileExit ? ':' : '.');
			//c = (tile == tileWall ? '#' : tile == tileBlock ? 'O' : '.');
			if (x == this->x && y == this->y)
				c = (dir == dirUp ? '^' : dir == dirDown ? 'v' : dir == dirLeft ? '<' : dir == dirRight ? '>' : '?');
			fputc(c, f);
		}
		fprintf(f, "\n");
	}
}

bool TPlayer::ApplyMove(TMove move, string &errorMsg)
{
	if (move == moveForward)
	{
		int xx = x + DX[dir], yy = y + DY[dir];
		TTile tile = board->GetTile(xx, yy);
		if (tile != tileClear && tile != tileExit) { errorMsg = "Zadel si se v zid."; return false; }
		x = xx; y = yy;
		if (carries) mc.nStepsWithBurden++; else mc.nStepsWoBurden++;
	}
	else if (move == moveTurnLeft || move == moveTurnRight || move == moveUTurn)
	{
		dir = (TDir) ((dir + (move == moveTurnLeft ? 1 : move == moveUTurn ? 2 : 3)) % 4);
		if (move == moveUTurn) {
			if (carries) mc.nUTurnsWithBurden++; else mc.nUTurnsWoBurden++; }
		else {
			if (carries) mc.nTurnsWithBurden++; else mc.nTurnsWoBurden++; }
	}
	else if (move == movePickUp)
	{
		if (carries) { errorMsg = "Ne mores pobrati kocke, ce eno ze nosis."; return false; }
		int xx = x + DX[dir], yy = y + DY[dir];
		TTile tile = board->GetTile(xx, yy);
		if (tile == tileClear || tile == tileExit) { errorMsg = "Ne mores pobrati kocke, ce je pred tabo prazno polje."; return false; }
		if (tile == tileWall) mc.nPickupAttempts++;
		else if (tile == tileBlock || tile == tileBlockExit)
		{
			carries = true;
			mc.nPickups++;
			board->SetTile(xx, yy, (tile == tileBlockExit ? tileExit : tileClear));
		}
		else assert(false); // unknown block type
	}
	else if (move == moveDrop)
	{
		if (! carries) { errorMsg = "Ne mores spustiti kocke, ce je ne nosis."; return false; }
		int xx = x + DX[dir], yy = y + DY[dir];
		TTile tile = board->GetTile(xx, yy);
		if (tile != tileClear && tile != tileExit) { errorMsg = "Ne mores odloziti kocke, ce polje pred tabo ni prazno."; return false; }
		mc.nDrops++; carries = false;
		board->SetTile(xx, yy, (tile == tileClear ? tileBlock : tileBlockExit));
	}
	else
		assert(false); // unknown move type

	return true;
}

bool TPlayer::PerformMove(TMove move, string &errorMsg)
{
	ApplyVisAreaToSeen(false);
	bool retVal = ApplyMove(move, errorMsg);
	RefreshVisibleArea();
	ApplyVisAreaToSeen(true);
	return retVal;
}

#ifdef TEST_MY

void TestMyBehavior(const char *fileName)
{
	TBoard board;
	if (fileName)
	{
		board = TBoard(fileName);
		if (board.xStart < 0 || board.yStart < 0) // select a random starting position
		{
			vector<int> pos;
			for (int y = 0; y < board.h; y++) for (int x = 0; x < board.w; x++)
			{
				TTile tile = board.GetTile(x, y);
				if (tile == tileClear) pos.push_back(y * board.w + x);
			}
			sort(pos.begin(), pos.end());
			int pos_ = pos[rand() % pos.size()];
			board.xStart = pos_ % board.w; board.yStart = pos_ / board.w;
			board.dirStart = TDir(rand() % 4);
		}
	}
	else
	{
		board = TBoard(10, 10, 10);
		board.BuildOuterWall();
		board.SetTile(9, 8, tileExit);
		board.SetTile(3, 0, tileWall);
		board.SetTile(4, 0, tileWall);
		board.SetTile(4, 1, tileWall);
		board.SetTile(2, 2, tileBlock);
		//board.SetTile(4, 2, tileExit);
	}
	TPlayer player(&board, board.vis, board.xStart, board.yStart, board.dirStart);
	//player.PrintVisArea();
	//player.PrintSeen();
	if (false)
	{
		string err;
		if (! player.PerformMove(moveForward, err)) printf("%s\n", err.c_str());
		if (! player.PerformMove(moveTurnLeft, err)) printf("%s\n", err.c_str());
		if (! player.PerformMove(movePickUp, err)) printf("%s\n", err.c_str());
		if (! player.PerformMove(moveTurnRight, err)) printf("%s\n", err.c_str());
		if (! player.PerformMove(moveUTurn, err)) printf("%s\n", err.c_str());
		if (! player.PerformMove(moveTurnRight, err)) printf("%s\n", err.c_str());
		if (! player.PerformMove(moveDrop, err)) printf("%s\n", err.c_str());
	}
	IBehavior *engine = CreateMyBehaviorInstance();
	FILE *f = fopen("gameLog.txt", "wt");
	player.RunGame(engine, f);
	fclose(f);

	if (false) {
		printf("\n\n");
		player.PrintVisArea();
		player.PrintSeen(); }
}

#endif // TEST_MY

//-----------------------------------------------------------------------------
// TServer
//-----------------------------------------------------------------------------

double GetCpuTimeSec(intptr_t pid)
{
#ifdef _WIN32
	if (pid == (intptr_t) -1) return 0;
	FILETIME ct, et, kt, ut;
	BOOL ok = GetProcessTimes((HANDLE) pid, &ct, &et, &kt, &ut);
	if (! ok) return -1;
	ULARGE_INTEGER uli, kli;
	uli.HighPart = ut.dwHighDateTime;
	uli.LowPart = ut.dwLowDateTime;
	kli.HighPart = kt.dwHighDateTime;
	kli.LowPart = kt.dwLowDateTime;
	return double(uli.QuadPart + kli.QuadPart) / 10000000.0;
#endif
}

string trim(const string s)
{
	size_t i = 0, j = s.size();
	while (i < j && isspace((unsigned char) s[i])) i++;
	while (i < j && isspace((unsigned char) s[j - 1])) j--;
	return s.substr(i, j - i);
}

class TMyException { };

class TServer
{
public:
	FILE *fOut, *fLog;
	SOCKET srvSocket, socket;
	TBoard *board;
	TPlayer *player;
	bool trace, drawBoards;
	string fnLog;
	intptr_t clientPid;
	int maxMoves, maxCpuSec;
	int portNo;
	TMoveCounts totalMoveCounts;
	int nWon, nTimeout, nWrongMove;
	double totalCpuSec;

	TServer()
	{
		fOut = stdout;
		srvSocket = INVALID_SOCKET;
		socket = INVALID_SOCKET;
		portNo = -1;
		board = 0; player = 0;
		trace = false;
		drawBoards = false;
		maxMoves = 1000;
		maxCpuSec = -1;
		clientPid = intptr_t(-1);
		fLog = 0; fnLog = "";
	}

	~TServer()
	{
		if (socket == INVALID_SOCKET) closesocket(socket);
		if (srvSocket == INVALID_SOCKET) closesocket(srvSocket);
		if (player) delete player;
		if (board) delete board;
		if (fOut != stdout && fOut != stderr) fclose(fOut);
	}

	void Send_(const char *s)
	{
		int len = (int) strlen(s), ofs = 0;
		while (ofs < len)
		{
			int nSent = send(socket, s + ofs, len - ofs, 0);
			if (nSent == SOCKET_ERROR) { fprintf(stderr, "Napaka pri posiljanju.\n"); throw TMyException(); }
			ofs += nSent;
		}
	}

	void SendLine(const char *s)
	{
		if (trace) { fprintf(fOut, ">> %s\n", trim(s).c_str()); }
		Send_(s);
	}

	void SendLine(const string &s)
	{
		SendLine(s.c_str());
	}

	string recvBuf;

	string ReadLine_()
	{
		string line = "";
		const char *p = recvBuf.c_str();
		if (p && *p) {
			const char *r = p; while (*r && *r != '\n') r++;
			if (*r == '\n') {
				r++; line = recvBuf.substr(0, r - p);
				recvBuf = recvBuf.substr(r - p); return line; }
			line = recvBuf; recvBuf = ""; }
		//
		enum { BufLen = 128 };
		char buf[BufLen + 1];
		while (true)
		{
			int nRead = recv(socket, buf, BufLen, 0);
			if (nRead == 0) break;
			if (nRead == SOCKET_ERROR) { fprintf(stderr, "Napaka pri branju.\n"); throw TMyException(); }
			int i = 0; while (i < nRead && buf[i] != '\n') i++;
			if (i >= nRead) { line.append(&buf[0], nRead); continue; }
			i++; recvBuf = ""; recvBuf.append(&buf[i], nRead - i);
			line.append(&buf[0], i); return line;
		}
		return "Stupid VC wants me to return something here.";
	}

	string ReadLine()
	{
		string s = ReadLine_();
		if (trace) { fprintf(fOut, "<< %s\n", trim(s).c_str()); }
		return s;
	}

	void DrawBoard(FILE *f)
	{
		if (! f) f = stdout;
		for (int y = 0; y < player->h; y++)
		{
			fprintf(f, "    ");
			for (int x = 0; x < player->w; x++)
			{
				char c = '.';
				TTile tile = player->board->GetTile(x, y);
				if (x == player->x && y == player->y)
					c = (player->dir == dirLeft ? '<' : player->dir == dirRight ? '>' : player->dir == dirUp ? '^' : 'v');
				else if (tile == tileBlockExit || tile == tileBlock) c = 'B';
				else if (tile == tileExit) c = 'x';
				else if (tile == tileWall) c = (x == 0 || y == 0 || x == player->w || y == player->h) ? 'Z' : '#';
				else if (tile == tileClear) c = '.';
				else c = '?';
				fputc(c, f);
			}
			fprintf(f, "\n");
		}
	}

	int PlayGame()
	{
		player = new TPlayer(board, board->vis, board->xStart, board->yStart, board->dirStart);
		char buf[100]; sprintf(buf, "ZBUDI_SE %d %d %d\r\n", board->w, board->h, board->vis);
		SendLine(buf);
		string buden = ReadLine();
		if (buden.substr(0, 5) != "BUDEN") { fprintf(stderr, "Pricakoval bi BUDEN, ne pa \"%s\".\n", trim(buden).c_str()); return 1; }
		int nMoves = 0; bool won = false, timeOut = false; string err = "";
		double tmStart = GetCpuTimeSec(clientPid), tmUsed = 0;
		while (nMoves < maxMoves)
		{
			sprintf(buf, "QV %s %d", (player->carries ? "B" : "N"), (int) player->visTiles.size());
			string qv = buf;
			for (int i = 0; i < (int) player->visTiles.size(); i++) {
				TVisibleTile vt = player->visTiles[i];
				sprintf(buf, " %d %d %s", vt.rowsInFront, vt.column,
					vt.what == visClear ? "." :
					vt.what == visBlockOrWall ? "#" :
					vt.what == visExit ? "x" :
					vt.what == visOuterWall ? "Z" : "?");
				qv.append(buf); }
			qv.append("\r\n");
			SendLine(qv);
			//
			string qvr = ReadLine();
			if (qvr.substr(0, 4) != "QVR ") { fprintf(stderr, "Pricakoval bi QVR, ne pa \"%s\".\n", trim(qvr).c_str()); return 1; }
			qvr = trim(qvr.substr(4));
			TMove move;
			if (0 == strcmpi(qvr.c_str(), "poberi")) move = movePickUp;
			else if (0 == strcmpi(qvr.c_str(), "spusti")) move = moveDrop;
			else if (0 == strcmpi(qvr.c_str(), "naprej")) move = moveForward;
			else if (0 == strcmpi(qvr.c_str(), "levo")) move = moveTurnLeft;
			else if (0 == strcmpi(qvr.c_str(), "desno")) move = moveTurnRight;
			else if (0 == strcmpi(qvr.c_str(), "uturn")) move = moveUTurn;
			else { fprintf(stderr, "Neveljaven premik: \"%s\".\n", qvr.c_str()); return 1; }
			//
			if (fLog) { fprintf(fLog, "Move %s\n", (move == moveForward ? "Forward" :
				move == moveTurnLeft ? "TurnLeft" :
				move == moveTurnRight ? "TurnRight" :
				move == moveUTurn ? "UTurn" :
				move == movePickUp ? "PickUp" :
				move == moveDrop ? "Drop" : "Invalid"));
				fflush(fLog); }
			//
			bool ok = player->PerformMove(move, err);
			if (! ok) break;
			nMoves++;
			printf("Poteza %d: %s.  Novi polozaj (%d, %d), smer %s.\n", nMoves, qvr.c_str(), player->x, player->y,
				(player->dir == dirUp ? "gor" : player->dir == dirLeft ? "levo" : player->dir == dirRight ? "desno" : "dol"));
			if (drawBoards) DrawBoard(stdout);
			if (clientPid != (intptr_t) -1) {
				double tmNow = GetCpuTimeSec(clientPid);
				tmUsed = tmNow - tmStart;
				if (maxCpuSec > 0 && tmUsed > maxCpuSec) { timeOut = true; break; }}
			if (player->board->GetTile(player->x, player->y) == tileExit) { won = true; break; }
		}
		if (won) err = "Cestitamo, prisel si do izhoda.";
		else if (nMoves >= maxMoves) err = "Naredil si prevec korakov.";
		else if (timeOut) err = "Porabil si prevec procesorjevega casa.";
		if (fLog)
		{
			fprintf(fLog, "%s %s\n", (won ? "Won" : "Error"), err.c_str()); fflush(fLog);
			fprintf(fLog, "Stats nStepsWoBurden %d\n", player->mc.nStepsWoBurden);
			fprintf(fLog, "Stats nStepsWithBurden %d\n", player->mc.nStepsWithBurden);
			fprintf(fLog, "Stats nTurnsWoBurden %d\n", player->mc.nTurnsWoBurden);
			fprintf(fLog, "Stats nTurnsWithBurden %d\n", player->mc.nTurnsWithBurden);
			fprintf(fLog, "Stats nUTurnsWoBurden %d\n", player->mc.nUTurnsWoBurden);
			fprintf(fLog, "Stats nUTurnsWithBurden %d\n", player->mc.nUTurnsWithBurden);
			fprintf(fLog, "Stats nPickups %d\n", player->mc.nPickups);
			fprintf(fLog, "Stats nPickupAttempts %d\n", player->mc.nPickupAttempts);
			fprintf(fLog, "Stats nDrops %d\n", player->mc.nDrops);
			fprintf(fLog, "Stats TotalMoves %d\n", player->mc.TotalMoves());
			fprintf(fLog, "Stats TotalCost %d\n", player->mc.TotalCost());
			if (clientPid != (intptr_t) -1) fprintf(fLog, "Stats CpuTimeSec %.3f\n", tmUsed);
		}
		//
		for (int i = 0; i < (int) err.size(); i++) if (isspace((unsigned char) err[i])) err[i] = ' ';
		SendLine(string("KONEC_IGRE ") + (won ? "ZMAGA" : "NAPAKA") + " " + err + "\r\n");
		//if (fLog)  { fprintf(fLog, "%s %s\n", (won ? "Won" : "Error"), err.c_str()); fflush(fLog); }
		totalMoveCounts += player->mc;
		if (won) nWon++; else if (nMoves >= maxMoves) nTimeout++; else nWrongMove++;
		//
		printf("Statistika po %d igrah: %d-krat uspesno prisel ven, %d-krat prevec potez, %d napak.\n",
			(nWon + nTimeout + nWrongMove), nWon, nTimeout, nWrongMove);
		printf("Stevilo potez (v zadnji igri in skupaj):\n");
		printf(" - Premik naprej: %d brez bremena, %d z  [skupaj %d brez, %d z]\n", player->mc.nStepsWoBurden, player->mc.nStepsWithBurden, totalMoveCounts.nStepsWoBurden, totalMoveCounts.nStepsWithBurden);
		printf(" - Zasuk: %d brez bremena, %d z  [skupaj %d brez, %d z]\n", player->mc.nTurnsWoBurden + player->mc.nUTurnsWoBurden, player->mc.nTurnsWithBurden + player->mc.nUTurnsWithBurden, totalMoveCounts.nTurnsWoBurden + totalMoveCounts.nUTurnsWoBurden, totalMoveCounts.nTurnsWithBurden + totalMoveCounts.nUTurnsWithBurden);
		printf(" - Poberi kocko: %d uspesno, %d ne (ker je zid) [skupaj %d uspesno, %d ne]\n", player->mc.nPickups, player->mc.nPickupAttempts, totalMoveCounts.nPickups, totalMoveCounts.nPickupAttempts);
		printf(" - Spusti kocko: %d [skupaj %d]\n", player->mc.nDrops, totalMoveCounts.nDrops);
		printf(" - Skupno stevilo potez: %d [skupaj %d]\n", player->mc.TotalMoves(), totalMoveCounts.TotalMoves());
		printf(" - Skupna cena potez: %d\n", player->mc.TotalCost());
		if (clientPid != (intptr_t) -1) { printf(" - Porabljeni CPU cas: %.3f s\n", tmUsed); totalCpuSec += tmUsed; }
		printf("\n");
		delete player; player = 0;
		return 0;
	}

	int main(TBoard *board_, int nGames)
	{
		srvSocket = ::socket(AF_INET, SOCK_STREAM, 0);
		if (srvSocket == INVALID_SOCKET) { fprintf(stderr, "Napaka pri ustvarjanju socketa.\n"); return 1; }
		sockaddr_in addr;
		addr.sin_family = AF_INET;
		addr.sin_addr.S_un.S_addr = inet_addr("127.0.0.1");
		addr.sin_port = htons(portNo);
		int ok = bind(srvSocket, (sockaddr*) &addr, sizeof(addr));
		if (ok != 0) { fprintf(stderr, "Napaka pri klicu bind().\n"); return 1; }
		fprintf(fOut, "Poslusam na portu %d.\n", portNo); fflush(fOut);
		ok = listen(srvSocket, SOMAXCONN);
		if (ok != 0) { fprintf(stderr, "Napaka pri klicu listen.\n"); return 1; }
		sockaddr_in clientAddr;
		int addrLen = (int) sizeof(clientAddr); socket = accept(srvSocket, (sockaddr*) &clientAddr, &addrLen);
		if (socket == INVALID_SOCKET) { fprintf(stderr, "Klic accept() je spodletel.\n"); return 1; }
		fprintf(fOut, "Povezan sem z %s:%d.\n", inet_ntoa(clientAddr.sin_addr), (int) ntohs(clientAddr.sin_port)); fflush(fOut);
		recvBuf = "";
		//
		SendLine("LabirintConsole 1.0\r\n");
		string ave = trim(ReadLine());
		if (ave.substr(0, 4) != "AVE ") { fprintf(stderr, "Priakoval bi AVE, ne pa \"%s\".\n", trim(ave).c_str()); return 1; }
		string playerName = ave.substr(4);
		SendLine("PRIJAVA USPESNA\r\n");
		//
		totalMoveCounts.Clr();
		nWon = 0; nTimeout = 0; nWrongMove = 0; totalCpuSec = 0;
		for (int gameNo = 0; nGames < 0 || gameNo < nGames; gameNo++)
		{
			board = new TBoard(*board_);
			fLog = 0;
			const char *p = fnLog.c_str(); if (strlen(p) > 0)
			{
				string fn = fnLog;
				const char *r = strstr(p, "%d");
				if (r) { char buf[10]; sprintf(buf, "%d", gameNo); fn = fnLog.substr(0, r - p); fn += buf; fn += fnLog.substr(r + 2 - p);  }
				fLog = fopen(fnLog.c_str(), "wt");
				board->Save(fLog);
				fprintf(fLog, "PlayerName %s\n", playerName.c_str()); fflush(fLog);
			}
			int retVal = PlayGame();
			if (fLog) { fclose(fLog); fLog = 0; }
			delete board; board = 0;
			if (retVal != 0) return retVal;
		}
		//
		SendLine("ADIJO\r\n");
		return 0;
	};
};

bool IsParam(const char *p, const char *name)
{
	if (p && p[0] == '-') {
		if (p[1] == '-' && strcmpi(p + 2, name) == 0) return true;
		if (strcmpi(p + 1, name) == 0) return true; }
	return false;
}

bool IsParamGetInt(int argc, char **argv, int& argNo, const char *name, int &value)
{
	if (argNo >= argc) return false;
	if (! IsParam(argv[argNo], name)) return false;
	if (argNo + 1 >= argc) { fprintf(stderr, "Napaka: za parametrom \"%s\" mora biti celo stevilo.\n", argv[argNo]); argNo++; return false; }
	int ok = sscanf(argv[argNo + 1], "%d", &value);
	if (ok != 1) { fprintf(stderr, "Napaka: za parametrom \"%s\" mora biti celo stevilo, ne pa \"%s\".\n", argv[argNo], argv[argNo + 1]); argNo += 2; return false; }
	argNo += 2; return true;
}

bool IsParamGetStr(int argc, char **argv, int& argNo, const char *name, string &value)
{
	if (argNo >= argc) return false;
	if (! IsParam(argv[argNo], name)) return false;
	if (argNo + 1 >= argc) { fprintf(stderr, "Napaka: za parametrom \"%s\" mora biti nek niz.\n", argv[argNo]); argNo++; return false; }
	value = argv[argNo + 1];
	argNo += 2; return true;
}

enum { DefaultPort = 1176 };

int ShowHelp()
{
	printf("Sintaksa: rtk09labirint parametri...\n\n");
	printf("Mozni parametri:\n");
	printf(" -lab filename.txt    Uporabi labirint iz datoteke filename.txt\n");
	printf(" -port 1234           Poslusaj na portu 1234 (default: %d)\n", (int) DefaultPort);
	printf(" -stIger 567          Odigraj 567 iger (default: 1)\n");
	printf(" -maxPotez 789        Dovoli najvec 789 potez na igro\n");
	printf(" -maxCpu 123          Dovoli klientu najvec 123 sekund CPU casa\n");
	printf(" -echo                Izpisuj mrezni promet tudi na zaslon\n");
	printf(" -slike               Po vsaki potezi narisi stanje labirinta\n");
	printf(" -log filename.txt    Shrani odigrano igro v filename.txt\n");
	printf("                        [ime lahko vsebuje %%d za stevilko igre]\n");
	printf(" -client mojRobot.exe tralala hopsasa   -- Pred zacetkom prve igre pozeni\n");
	printf("          program mojRobot.exe s parametroma \'tralala\' in \'hopsasa\'.\n\n");
	printf("\'-lab\' je obvezen.  \'-client\', ce je prisoten, mora biti zadnji.\n");
	printf("\'-maxCpu\' je dovoljen le, ce je podan tudi \'-client\'.\n");
	return 2;
}

int ServerMain(int argc, char **argv)
{
	printf("Roboti v labirintu - testni streznik (verzija 1.1)\n");
	for (int i = 0; i < argc; i++) {
		const char *p = argv[i];
		if (p[0] == '?') return ShowHelp();
		if (! (p[0] == '/' || p[0] == '-')) continue;
		if (strcmpi(p + 1, "h") == 0 || strcmpi(p + 1, "help") == 0 || strcmpi(p + 1, "?") == 0 || strcmpi(p + 1, "-help") == 0) return ShowHelp(); }
	//
	TServer server;
	server.portNo = DefaultPort;
	int nGames = 1;
	TBoard *board = 0; string fnBoard, fnLog;
	int iClient = -1; string fnClient;
	bool showHelp = false;
	for (int i = 1; i < argc; )
	{
		if (IsParamGetInt(argc, argv, i, "port", server.portNo)) continue;
		if (IsParamGetInt(argc, argv, i, "stIger", nGames)) continue;
		if (IsParamGetInt(argc, argv, i, "maxPotez", server.maxMoves)) continue;
		if (IsParamGetInt(argc, argv, i, "maxCpu", server.maxCpuSec)) continue;
		if (IsParamGetStr(argc, argv, i, "lab", fnBoard))
		{
			if (! board) delete board;
			board = new TBoard(fnBoard.c_str());
			if (board->xStart < 0 || board->yStart < 0) // start at a random position
			{
				vector<int> cands; for (int cand = 0; cand < board->w * board->h; cand++) if (board->GetTile(cand % board->w, cand / board->w) == tileClear) cands.push_back(cand);
				int pos = cands[rand() % int(cands.size())];
				board->xStart = pos % board->w; board->yStart = pos / board->w;
				board->dirStart = TDir(rand() % 4);
				printf("V opisu labirinta ni bil oznaen zaetni poloaj robota, zato sem ga izbral nakljuno: (%d, %d), smer %s.\n",
					board->xStart, board->yStart, (board->dirStart == dirLeft ? "<" : board->dirStart == dirRight ? ">" : board->dirStart == dirUp ? "^" : board->dirStart == dirDown ? "v" : "?"));
			}
			continue;
		}
		if (IsParam(argv[i], "echo")) { server.trace = true; i++; continue; }
		if (IsParam(argv[i], "slike")) { server.drawBoards = true; i++; continue; }
		if (IsParamGetStr(argc, argv, i, "log", server.fnLog)) continue;
		if (IsParamGetStr(argc, argv, i, "client", fnClient)) { iClient = i - 1; break; }
		fprintf(stderr, "Neznan parameter: \"%s\"\n", argv[i]); showHelp = true; i++;
	}
	if (!board) { fprintf(stderr, "Napaka: manjka parameter -lab, ki pove ime datoteke z labirintom.\n\n"); ShowHelp(); return 1; }
	if (showHelp) ShowHelp();
	if (iClient >= 0)
	{
		printf("Poganjam \"%s\"...\n", fnClient.c_str());
		intptr_t ok = _spawnvp(_P_NOWAIT, fnClient.c_str(), argv + iClient);
		if (ok == intptr_t(-1)) printf("Napaka pri poganjanju klienta: %d (%s)\n", errno, strerror(errno));
		server.clientPid = ok;
	}
#ifdef _WIN32
	WSADATA wsaData; if (WSAStartup(0x101, &wsaData) != 0) { fprintf(stderr, "Napaka pri klicu WSAStartup.\n"); return 1; }
#endif
	int retVal = server.main(board, nGames);
	if (retVal != 0) fprintf(stderr, "Stevilka napake: %d\n", WSAGetLastError());
#ifdef _WIN32
	WSACleanup();
#endif
	return retVal;
}

#ifndef CLIENT
int main(int argc, char **argv)
{
#ifdef TEST_MY
	return TestMyBehavior(argc > 1 ? argv[1] : 0);
#else
	return ServerMain(argc, argv);
#endif
	return 0;
}
#endif

// -lab lab00.txt -echo -slike -log myLog.txt -maxPotez 100
