#ifndef ____Rtk09Labirint_defined____
#define ____Rtk09Labirint_defined____

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <vector>
#include <string>
#include <cassert>

//-----------------------------------------------------------------------------
// TVisibleTile
//-----------------------------------------------------------------------------
// A list of these tiles is reported to the player before making a move.

enum { visClear = '.', visExit = 'x', visBlockOrWall = '#', visUnknown = '?', visOuterWall = 'Z' };

class TVisibleTile
{
public:
	char what;
	int rowsInFront; // 1..vis
	int column; // -rowsInFront..rowsInFront

	TVisibleTile(int x, int y, char what_) : what(what_), column(x), rowsInFront(y) {
		assert(y > 0); assert(-y <= x && x <= y); }
};

//-----------------------------------------------------------------------------
// Limits and constants
//-----------------------------------------------------------------------------

enum { MaxW = 100, MaxH = 100, MaxVis = 100 };

class MoveCost
{
public:
	enum 
	{ 
		WalkB = 2, WalkNB = 1, // with/without burden
		TurnB = 2, TurnNB = 1, // with/without burden; turns for 90 degrees
		UTurnB = TurnB, UTurnNB = TurnNB, // with/without burden; turn for 180 degrees
		PickUp = 2, TryPickup = 1, // TryPickup = if you try to pick up a wall
		Drop = 1
	};
};

//-----------------------------------------------------------------------------
// IBehavior
//-----------------------------------------------------------------------------
// Must be implemented to define the player's behaviour.

typedef enum TMove_ 
{ 
	moveForward = 0, moveTurnLeft = 1, moveTurnRight = 2, 
	moveUTurn = 3, movePickUp = 4, moveDrop = 5, MoveMax = 6 
} TMove;

class IBehavior
{
public:
	virtual ~IBehavior() { }
	virtual std::string GetName() = 0;
	// Tells the player that a new game is beginning, with a 
	// board of the given size.
	virtual void BeginGame(int w, int h, int vis) = 0;
	// 'carries' tells if the player is currently carrying a block;
	// 'visTiles' reports the tiles he currently sees, relatively
	// to his current position and orientation.
	virtual TMove NextMove(bool carries, std::vector<TVisibleTile> visTiles) = 0;
	// Tells the player that the game has ended.
	virtual void EndGame(bool won, std::string reason) = 0;
};

inline const char *MoveStr(TMove move)
{
	if (move == moveForward) return "Forward";
	else if (move == moveTurnLeft) return "Left";
	else if (move == moveTurnRight) return "Right";
	else if (move == movePickUp) return "PickUp";
	else if (move == moveDrop) return "Drop";
	else if (move == moveUTurn) return "UTurn";
	else { assert(false); return "???"; }
}

//-----------------------------------------------------------------------------
// Timer
//-----------------------------------------------------------------------------

class TTimer
{
protected:
	__int64 count, start; int running;
	static __int64 Clock() {
		FILETIME ct, et, kt, ut;
		ULARGE_INTEGER ul;
		GetProcessTimes(GetCurrentProcess(), &ct, &et, &kt, &ut);
		ul.LowPart = ut.dwLowDateTime; ul.HighPart = ut.dwHighDateTime;
		return ul.QuadPart; }
public:
	TTimer() : running(0), count(0), start(0) { }
	void Start() { if (running++ == 0) start = Clock(); }
	TTimer& Stop() { if (--running == 0) count += Clock() - start; return *this; }
	__int64 Count() const { __int64 retVal = count; if (running > 0) retVal += Clock() - start; return retVal; }
	double Sec() const { return Count() / 10000000.0; }
	double Ms() const { return Count() / 10000.0; }
};

#endif // ____Rtk09Labirint_defined____